// Server module
// By Adam Chlipala

#include "3d.h"

char buf[1024], buf2[1024], buf3[1024], txt[1024];

int n, id;
matrix m;
vector v;

void sync(void)
{
	char buf[1024];

	GetName(buf);
	SetName(buf);
	SetMass(GetMass());
	SetElasticity(GetElasticity());
	if (GetWalk())
		SetWalk(GetWalk());
	SetStationary(GetStationary());
	SetGlobal(GetGlobal());
	
	LocalSelectComponent(0);
	SelectComponent(0);

	GetTrans(v);
	SetTrans(v);
	GetVel(v);
	SetVel(v);
	GetRotVel(m);
	SetRotVel(m);
	GetRV(v);
	SetRV(v);
	GetRot(m);
	SetRot(m);
}

void main(int e, int a, int a2, int a3)
{
	switch (e)
	{
	case EVENT_INIT:
		puts("Server module init....");

		AllocSmem(8);
		n = 0;
		SetSmem(&n, 0, 4);
		SetSmem(&n, 4, 4);

		SetWorldInfo(2000, 400, 2000, 200);
		AddUser("root", "root", 256);

		OpenEnt(0, ACC_WRITE);
		SelectEnt(0);
		AddGravity(0, -0.1, 0);
		CloseEnt(0);

		n = AddEntity(0, "base_cmds.xc", 0, -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SetStationary(STAT_ABSORB);
		CloseEnt(n);

		n = AddEntity("rsf.xc", 0, "ground.rsf", -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetGlobal(1);
		SetMass(100);
		SetName("ground");
		SetElasticity(0.01);
		SetStationary(STAT_ABSORB);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("rsf.xc", NULL, "sky.rsf", -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetMass(1000);
		SetGlobal(1);
		SetName("The Wild Blue Yonder");
		SetStationary(STAT_ABSORB);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("rsf.xc", 0, "hut.rsf", -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetStationary(STAT_ABSORB);
		SetMass(20);
		SetName("Hut");
		v[0] = -35;
		v[1] = 4.38733;
		v[2] = 0;
		SetTrans(v);
		RotationMatrix(m, 0, 1, 0, 90);
		SetRot(m);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("rsf.xc", 0, "hut.rsf", -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetStationary(STAT_ABSORB);
		SetMass(20);
		SetName("Hut");
		v[0] = 35;
		v[1] = 4.38733;
		v[2] = 0;
		SetTrans(v);
		RotationMatrix(m, 0, 1, 0, -90);
		SetRot(m);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("ballr.xc", "ball.xc", 0, -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetMass(5);
		SetElasticity(0);
		SetName("GeigerBall!");
		v[0] = -41;
		v[1] = 5;
		v[2] = 0;
		SetTrans(v);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("plane.xc", 0, 0, -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetActive(1);
		CloseEnt(n);

		n = AddEntity("ttt.xc", "ttt_con.xc", 0, -1);
		OpenEnt(n, ACC_WRITE);
		SelectEnt(n);
		SelectComponent(0);
		SetStationary(STAT_ABSORB);
		SetName("Tic-Tac-Toe");
		v[0] = v[2] = 0;
		v[1] = 1.77812;
		SetTrans(v);
		RotationMatrix(m, 1, 0, 0, -30);
		SetRot(m);
		SetActive(1);
		CloseEnt(n);

		puts("Done");

		break;
	case EVENT_SHUTDOWN:
		puts("Server module shutting down");
		break;
	case EVENT_LOGON:
		GetStringArg(buf);
		GetStringArg2(buf2);

		if (!strcmp(buf, "guest"))
		{
			GetSmem(&n, 4, 4);
			++n;
			SetSmem(&n, 4, 4);

			sprintf(buf, "guest%d", n);
			n = AddUser(buf, "TEMP", 0);
		}
		else
		{
			n = GetUserID(buf);
			if (n < 0)
			{
				Quit("User not found");
				break;
			}

			GetPass(n, buf3);
			if (strcmp(buf2, buf3))
			{
				Quit("Incorrect password");
				break;
			}

			if (GetUserAvatar(n))
			{
				Quit("Multiple connections per account not allowed");
				break;
			}
		}

		SetUser(n);
	
		id = AddEntity("rsf.xc", 0, "george.rsf", -1);
		OpenEnt(id, ACC_WRITE);
		SelectEnt(id);
		SetOwner(n);
		GetSmem(&n, 0, sizeof(int));

		n %= 4;
		v[0] = v[2] = 0;
		v[1] = 40;
		SelectComponent(0);
		switch (++n)
		{
			case 1:
				v[2] = 20;
				SetTrans(v);
				break;
			case 2:
				v[2] = -20;
				SetTrans(v);
				RotationMatrix(m, 0, 1, 0, 180);
				SetRot(m);
				break;
			case 3:
				v[0] = -20;
				SetTrans(v);
				RotationMatrix(m, 0, 1, 0, 90);
				SetRot(m);
				break;
			case 4:
				v[0] = 20;
				SetTrans(v);
				RotationMatrix(m, 0, 1, 0, 270);
				SetRot(m);
				break;
		}

		SetName(buf);
		strcat(buf, " has connected.");
		OpenEnt(0, ACC_WRITE);
		SelectEnt(0);
		ZoneText(buf);
		CloseEnt(0);

		SelectEnt(id);
		SetAvatar(id);
		SetMass(10);
		SetActive(1);
		CalcZone();
		CloseEnt(id);
		SetSmem(&n, 0, sizeof(int));

		break;
	case EVENT_OLOGON:
/*		GetStringArg(buf);
		GetStringArg2(buf2);

		n = GetUserID(buf);
		if (n < 0)
		{
			Quit("User not found");
			break;
		}
		GetPass(n, buf3);
		if (strcmp(buf2, buf3))
		{
			Quit("Incorrect password");
			break;
		}

		SetOAvatar(GetUserAvatar(n));*/
		break;
	case EVENT_REMOTE_INIT:
		WriteText("Welcome to the [Unnamed Multi-User Networked 3D Engine] test server. To move around click outside of any of the floating windows, then use up/down to move forward/backward, left/right to spin, and home/end to look up/down. Click various objects and you might be pleasantly surprised! Click back in the text entry box at the bottom of the screen to type messages or /commands.");
		TimeRef(time(0));

		SetWorldInfo(GetWorldX(), GetWorldY(), GetWorldZ(), GetVZoneSize());
		OpenEnt(0, ACC_WRITE);
		SelectEnt(0);
		AddGravity(0, -0.1, 0);
		CloseEnt(0);

		break;
	case EVENT_PRE_SYNC:
		SetFreeze(1);
		break;
	case EVENT_POST_SYNC:
		SetFreeze(0);
		break;
	case EVENT_CREATE:
		LocalOpenEnt(a, ACC_READ);
		LocalSelectEnt(a);

		GetRenderMod(buf);
		GetMod(buf2);
		GetArgs(buf3);

		AddEntity(buf, buf2, buf3, a);

		OpenEnt(a, ACC_WRITE);
		SelectEnt(a);

		GetName(buf);
		SetName(buf);
		SetMass(GetMass());
		SetElasticity(GetElasticity());
		if (GetWalk())
			SetWalk(GetWalk());
		SetStationary(GetStationary());
		SetGlobal(GetGlobal());
		
		LocalSelectComponent(0);
		SelectComponent(0);

		GetTrans(v);
		SetTrans(v);
		GetVel(v);
		SetVel(v);
		GetRotVel(m);
		SetRotVel(m);
		GetRV(v);
		SetRV(v);
		GetRot(m);
		SetRot(m);

		SetActive(GetActive());
		CloseEnt(a);
		
		LocalCloseEnt(a);

		break;
	case EVENT_POST_LOGON:
		SetAvatar(GetAvatar());

		break;
	case EVENT_WRITE_TEXT:
		GetStringArg(buf);
		WriteText(buf);

		break;
	case EVENT_SYNC:
		LocalOpenEnt(a, ACC_READ);
		LocalSelectEnt(a);
		OpenEnt(a, ACC_WRITE);
		SelectEnt(a);

		LocalSelectComponent(0);
		SelectComponent(0);

		if (a2 == UPD_ALL)
		{
			GetName(buf);
			SetName(buf);
			SetMass(GetMass());
			SetElasticity(GetElasticity());
			SetStationary(GetStationary());
			SetGlobal(GetGlobal());
		}

		if (a2 & UPD_WALK)
		{
			SetWalk(GetWalk());
		}

		if (a2 & UPD_POS)
		{
			GetTrans(v);
			SetTrans(v);
		}

		if (a2 & UPD_VEL)
		{
			GetVel(v);
			SetVel(v);
		}

		if (a2 & UPD_RVEL)
		{
			GetRotVel(m);
			SetRotVel(m);
		}

		if (a2 & UPD_RV)
		{
			GetRV(v);
			SetRV(v);
		}

		if (a2 & UPD_ROT)
		{
			GetRot(m);
			SetRot(m);
		}

		CloseEnt(a);
		LocalCloseEnt(a);

		break;
	case EVENT_CLOSE:
		OpenEnt(a, ACC_WRITE);
		SelectEnt(a);
		RemoveEntity();

		break;
	case EVENT_MODEL:
		OpenEnt(a, ACC_WRITE);
		SelectEnt(a);
		ModelEvent(a2, a3);
		CloseEnt(a);

		break;
	case EVENT_QUIT:
		GetStringArg(buf);
		Quit(buf);
		Sync();

		break;
	case EVENT_LOGOFF:
		OpenEnt(a, ACC_PUBLIC);
		SelectEnt(a);
		GetName(buf);
		strcat(buf, " has disconnected.");
		CloseEnt(a);

		OpenEnt(0, ACC_WRITE);
		SelectEnt(0);
		ZoneText(buf);
		CloseEnt(0);

		break;
	}
}
