// Tic-Tac-Toe control module
// By Adam Chlipala

#include "3d.h"

#define CLICK 100
#define DRAW_X_BLUE 110
#define DRAW_X_RED 111
#define CLEAR 112
#define NEWGAME 113
#define HELP 114
#define JOIN 115
#define QUIT 116

#define SGI SetCmem(&gameinfo, 0, sizeof gi)
#define GGI GetCmem(&gameinfo, 0, sizeof gi)

int x, y, j, k;

typedef struct
{
	char board[3][3], turn, name[2][80], n_s;
	int ent[2], score[2], games, ties;
} gi;
gi gameinfo;

char buf[256];

void newgame(void)
{
	for (j = 0; j < 3; ++j)
		for (k = 0; k < 3; ++k)
			gameinfo.board[j][k] = 0;

	gameinfo.turn = gameinfo.n_s = 0;

	SGI;

	ModelEvent(CLEAR, 0);

	OpenEnt(gameinfo.ent[0], ACC_PUBLIC);
	SelectEnt(gameinfo.ent[0]);
	WriteMsg("A new Tic-Tac-Toe game has begun. You are red. It is your turn.");
	CloseEnt(gameinfo.ent[0]);

	OpenEnt(gameinfo.ent[1], ACC_PUBLIC);
	SelectEnt(gameinfo.ent[1]);
	WriteMsg("A new Tic-Tac-Toe game has begun. You are blue. You will go second.");
	CloseEnt(gameinfo.ent[1]);
}

void init(void)
{
	for (j = 0; j < 3; ++j)
		for (k = 0; k < 3; ++k)
			gameinfo.board[j][k] = 0;

	gameinfo.turn = 2;
	gameinfo.ent[0] = gameinfo.ent[1] = 0;
}

void main(int event, int arg, int arg2, int arg3)
{
	switch (event)
	{
	case EVENT_INIT:
		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);

		AllocCmem(sizeof gi);
		SetElasticity(0);
		BindClick(CLICK);
		BindCmd("tttnew", NEWGAME);
		BindCmd("ttthelp", HELP);
		BindCmd("tttjoin", JOIN);
		BindCmd("tttquit", QUIT);

		init();
		SGI;

		CloseEnt(arg);

		break;
	case EVENT_SYNC:
		LocalOpenEnt(arg, ACC_WRITE);
		LocalSelectEnt(arg);
		GGI;

		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);

		ModelEvent(CLEAR, 0);

		for (x = 0; x < 3; ++x)
			for (y = 0; y < 3; ++y)
				if (gameinfo.board[x][y] == 1)
					ModelEvent(DRAW_X_RED, x*3 + y + 12);
				else if (gameinfo.board[x][y] == 2)
					ModelEvent(DRAW_X_BLUE, x*3 + y + 12);

		CloseEnt(arg);

		LocalCloseEnt(arg);

		break;
	case CLICK:
		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);
		GGI;

		if (gameinfo.turn == 2)
			break;
		if (arg2 == gameinfo.ent[gameinfo.turn] && arg3 >= 3 && arg3 < 12)
		{
			arg3 -= 3;
			x = arg3 / 3;
			y = arg3 % 3;

			if (!gameinfo.board[x][y])
			{
				gameinfo.board[x][y] = gameinfo.turn+1;

				if (gameinfo.turn)
					ModelEvent(DRAW_X_BLUE, arg3+12);
				else
					ModelEvent(DRAW_X_RED, arg3+12);

				x = gameinfo.turn+1;
				y = 0;

				for (j = 0; j < 3; ++j)
					if (gameinfo.board[0][j] == x && gameinfo.board[1][j] == x && gameinfo.board[2][j] == x)
					{
						y = 1;
						break;
					}
				if (!y)
					for (j = 0; j < 3; ++j)
						if (gameinfo.board[j][0] == x && gameinfo.board[j][1] == x && gameinfo.board[j][2] == x)
						{
							y = 1;
							break;
						}
				if (!y)
					if ((gameinfo.board[0][0] == x && gameinfo.board[1][1] == x && gameinfo.board[2][2] == x) || (gameinfo.board[2][0] == x && gameinfo.board[1][1] == x && gameinfo.board[0][2] == x))
						y = 1;

				if (y)
				{
					OpenEnt(gameinfo.ent[gameinfo.turn], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[gameinfo.turn]);
					WriteMsg("The Tic-Tac-Toe game is over. You won!");

					OpenEnt(gameinfo.ent[!gameinfo.turn], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[!gameinfo.turn]);
					WriteMsg("The Tic-Tac-Toe game is over. You lost.");

					++gameinfo.score[gameinfo.turn];
					++gameinfo.games;

					sprintf(buf, "The record so far: %s: %d, %s: %d, Ties: %d",
						gameinfo.name[0], gameinfo.score[0],
						gameinfo.name[1], gameinfo.score[1], gameinfo.ties);
					SelectEnt(gameinfo.ent[gameinfo.turn]);
					WriteMsg(buf);
					SelectEnt(gameinfo.ent[!gameinfo.turn]);
					WriteMsg(buf);

					CloseEnt(gameinfo.ent[gameinfo.turn]);
					CloseEnt(gameinfo.ent[!gameinfo.turn]);

					gameinfo.turn = 2;
				}
				else if (++gameinfo.n_s >= 9)
				{
					OpenEnt(gameinfo.ent[gameinfo.turn], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[gameinfo.turn]);
					WriteMsg("The Tic-Tac-Toe game is over. It is a tie.");

					OpenEnt(gameinfo.ent[!gameinfo.turn], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[!gameinfo.turn]);
					WriteMsg("The Tic-Tac-Toe game is over. It is a tie.");

					++gameinfo.ties;
					++gameinfo.games;

					sprintf(buf, "The record so far: %s: %d, %s: %d, Ties: %d",
						gameinfo.name[0], gameinfo.score[0],
						gameinfo.name[1], gameinfo.score[1], gameinfo.ties);
					SelectEnt(gameinfo.ent[gameinfo.turn]);
					WriteMsg(buf);
					SelectEnt(gameinfo.ent[!gameinfo.turn]);
					WriteMsg(buf);

					CloseEnt(gameinfo.ent[gameinfo.turn]);
					CloseEnt(gameinfo.ent[!gameinfo.turn]);

					gameinfo.turn = 2;
				}
				else
				{
					gameinfo.turn = !gameinfo.turn;

					OpenEnt(gameinfo.ent[gameinfo.turn], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[gameinfo.turn]);
					WriteMsg("It is your turn.");
					CloseEnt(gameinfo.ent[gameinfo.turn]);
				}

				SelectEnt(arg);
				SGI;
			}
		}

		break;
	case NEWGAME:
		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);
		GGI;

		if (arg2 == gameinfo.ent[0] || arg2 == gameinfo.ent[1])
		{
			if (!gameinfo.ent[0] || !gameinfo.ent[1])
			{
				OpenEnt(arg2, ACC_PUBLIC);
				SelectEnt(arg2);
				WriteMsg("You need one more player.");
				CloseEnt(arg2);
			}
			else
				newgame();
		}
		else
		{
			OpenEnt(arg2, ACC_PUBLIC);
			SelectEnt(arg2);
			WriteMsg("You are not in the game.");
			CloseEnt(arg2);
		}

		CloseEnt(arg);

		break;
	case HELP:
		OpenEnt(arg2, ACC_PUBLIC);
		SelectEnt(arg2);

		WriteMsg("The magic Tic-Tac-Toe board! To join a game when one is not in process, type /tttjoin. Simply click a square to make your move when prompted to do so. Start a new game by typing /tttnew. Leave play by typing /tttquit.");

		CloseEnt(arg2);

		break;
	case JOIN:
		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);
		GGI;

		OpenEnt(arg2, ACC_PUBLIC);
		SelectEnt(arg2);

		for (j = 0; j < 2; ++j)
		{
			if (!gameinfo.ent[j])
			{
				gameinfo.ent[j] = arg2;
				GetName(gameinfo.name[j]);
				gameinfo.score[j] = 0;

				WriteMsg("Welcome to the Tic-Tac-Toe game! Type /ttthelp for help.");

				if (gameinfo.ent[0] && gameinfo.ent[1])
				{
					gameinfo.ties = 0;

					WriteMsg("You may begin by typing /tttnew.");

					OpenEnt(gameinfo.ent[!j], ACC_PUBLIC);
					SelectEnt(gameinfo.ent[!j]);
					sprintf(buf, "%s has joined. You may begin by typing /tttnew.", gameinfo.name[j]);
					WriteMsg(buf);
				}

				SelectEnt(arg);
				SGI;
				CloseEnt(arg2);

				return;
			}
		}

		WriteMsg("Sorry, the game is already full! You'll have to watch.");

		return;
	case QUIT:
		OpenEnt(arg, ACC_WRITE);
		SelectEnt(arg);
		GGI;

		if (arg2 == gameinfo.ent[0] || arg2 == gameinfo.ent[1])
		{
			j = (arg2 == gameinfo.ent[1]);

			OpenEnt(arg2, ACC_PUBLIC);
			SelectEnt(arg2);
			WriteMsg("Adios. Thanks for playing.");
			CloseEnt(arg2);

			if (gameinfo.ent[!j])
			{
				OpenEnt(gameinfo.ent[!j], ACC_PUBLIC);
				SelectEnt(gameinfo.ent[!j]);
				sprintf(buf, "%s has left the Tic-Tac-Toe game.", gameinfo.name[j]);
				WriteMsg(buf);
				CloseEnt(gameinfo.ent[!j]);
			}

			gameinfo.ent[j] = 0;
			gameinfo.turn = 2;
			SelectEnt(arg);
			SGI;
		}
		else
		{
			OpenEnt(arg2, ACC_PUBLIC);
			SelectEnt(arg2);
			WriteMsg("You are not in the game.");
			CloseEnt(arg2);
		}

		CloseEnt(arg);

		break;
	}
}